import keyboard  # https://pypi.org/project/keyboard/
import platform


class Input:
    def __init__(self) -> None:
        self.receiving_input = True
        self.lines = []

    def get_lines(self) -> str:
        keyboard.add_hotkey("ctrl+enter", self._toggle_receiving_input)
        while self.receiving_input:
            self.lines.append(input())
        return "\n".join(self.lines)

    def _toggle_receiving_input(self) -> None:
        keyboard.write("\n")  # Some terminals require this for `input` to return.
        self.receiving_input = not self.receiving_input


def get_lines() -> str:
    print(end="\x1b[90m(")
    if platform == "darwin":
        print(end="cmd")
    else:
        print(end="ctrl")
    print("+enter to submit)\x1b[39m")
    return Input().get_lines()



from typing import Dict
from typing import List
import sqlite3


async def create_aliases_table(db_file_name: str) -> Dict[str, str]:
    """Creates a sqlite table with default aliases and returns all aliases.

    Assumes the table does not exist.
    """
    default_aliases = {
        "c": "c-clang",
        "c#": "cs-csc",
        "c++": "cpp-clang",
        "cpp": "cpp-clang",
        "cs": "cs-csc",
        "f#": "fs-core",
        "fs": "fs-core",
        "java": "java-openjdk",
        "javascript": "javascript-node",
        "js": "javascript-node",
        "objective-c": "objective-c-clang",
        "py": "python3",
        "python": "python3",
        "swift": "swift4",
    }
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            CREATE TABLE aliases (
                id INTEGER PRIMARY KEY,
                alias TEXT NOT NULL,
                language TEXT NOT NULL,
                UNIQUE (alias)
            );
            """
        )
        cursor.executemany(
            """
            INSERT OR IGNORE INTO aliases
            (alias, language)
            VALUES (?, ?);
            """,
            default_aliases.items(),
        )
        conn.commit()
    return default_aliases


async def load_aliases(db_file_name: str) -> Dict[str, str]:
    """Loads aliases from the database.

    Returns an empty dictionary if there are no aliases.
    """
    try:
        with sqlite3.connect(db_file_name) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT alias, language FROM aliases")
            records = cursor.fetchall()
            if not records:
                return dict()
            all_aliases = dict()
            for alias, language in records:
                all_aliases[alias] = language
            return all_aliases
    except sqlite3.OperationalError:
        return await create_aliases_table(db_file_name)


async def create_alias(
    db_file_name: str,
    new_alias: str,
    language: str,
    aliases: Dict[str, str],
    languages: List[str],
) -> None:
    aliases[new_alias] = language
    languages.append(new_alias)
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            INSERT INTO languages
            (language)
            VALUES (?);
            """,
            [new_alias],
        )
        cursor.execute(
            """
            INSERT INTO aliases
            (alias, language)
            VALUES (?, ?);
            """,
            (new_alias, language),
        )


async def delete_alias(
    alias: str, aliases: Dict[str, str], languages: List[str], db_file_name: str
):
    del aliases[alias]
    languages.remove(alias)
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            DELETE FROM aliases
            WHERE alias = ?;
            """,
            [alias],
        )
        cursor.execute(
            """
            DELETE FROM languages
            WHERE language = ?;
            """,
            [alias],
        )
        cursor.execute(
            """
            DELETE FROM jargon
            WHERE language = ?;
            """,
            [alias],
        )



from aliases import create_alias
from aliases import delete_alias
from aliases import load_aliases
from errors import InputError
from jargon import create_jargon
from jargon import delete_jargon
from jargon import has_jargon
from jargon import init_jargon
from jargon import print_jargon
from jargon import wrap_jargon
from multiline_input import get_lines
from textwrap import dedent
from typing import Dict
from typing import List
from typing import Tuple
import aiohttp  # https://docs.aiohttp.org/en/stable/
import argparse
import async_tio  # https://pypi.org/project/async-tio/
import asyncio
import sqlite3
import sys


VERSION = "0.4.2"


def init_argparse() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        usage="%(prog)s [OPTION]",
        description="Quickly run code in almost any language.",
        epilog="For tips, the source code, discussions, and more, visit https://github.com/wheelercj/tias",
    )
    parser.add_argument("-v", "--version", action="version", version=f"v{VERSION}")
    return parser


def main() -> None:
    _ = init_argparse().parse_args()
    loop = asyncio.get_event_loop()
    loop.run_until_complete(amain(loop))


async def amain(loop) -> None:
    connector = aiohttp.TCPConnector(force_close=True)
    async with aiohttp.ClientSession(connector=connector) as session:
        db_file_name = "tias.db"
        aliases: Dict[str, str] = await load_aliases(db_file_name)
        languages: List[str] = await load_languages(
            loop, session, db_file_name, aliases
        )
        await init_jargon(db_file_name)
        while True:
            try:
                choice = input("\x1b[32mtias> \x1b[39m").lower().strip()
                await parse_choice(
                    loop, session, db_file_name, languages, aliases, choice
                )
            except InputError as e:
                print(e)


async def parse_choice(
    loop,
    session,
    db_file_name: str,
    languages: List[str],
    aliases: Dict[str, str],
    choice: str,
) -> None:
    if choice == "help":
        await print_help()
    elif choice == "exit":
        sys.exit(0)
    elif choice == "":
        return
    elif choice.startswith("run "):
        language = choice.replace("run ", "").strip()
        if language not in languages:
            raise InputError(f"Invalid language: `{language}`")
        language, code, inputs = await get_code(language, aliases, db_file_name)
        if not code:
            raise InputError("Error: no code given to run.")
        await run_code(
            loop,
            session,
            db_file_name,
            languages,
            aliases,
            language,
            code,
            inputs,
        )
    elif choice == "list" or choice.startswith("list "):
        filter_keyword = ""
        if choice.startswith("list "):
            filter_keyword = choice.replace("list ", "").strip()
        await list_languages(languages, aliases, filter_keyword)
    elif choice.startswith("jargon "):
        language = choice.replace("jargon ", "").strip()
        if language not in languages:
            raise InputError(f"Invalid language: `{language}`")
        await print_jargon(language, db_file_name)
    elif choice.startswith("create jargon "):
        language = choice.replace("create jargon ", "").strip()
        if language not in languages:
            raise InputError(f"Invalid language: `{language}`")
        if await has_jargon(language, db_file_name):
            c = input(f"`{language}` already has jargon. Overwrite? (y/n) ")
            if c.lower().strip() not in ("y", "yes"):
                raise InputError("Cancelled creating jargon.")
            await delete_jargon(language, db_file_name)
        await create_jargon(language, db_file_name)
        print(f"Created jargon for the `{language}` language")
    elif choice.startswith("delete jargon "):
        language = choice.replace("delete jargon ", "").strip()
        if language not in languages:
            raise InputError(f"Invalid language: `{language}`")
        if not await has_jargon(language, db_file_name):
            raise InputError(f"`{language}` has no jargon")
        await delete_jargon(language, db_file_name)
        print(f"Jargon for the `{language}` language deleted.")
    elif choice.startswith("alias "):
        alias = choice.replace("alias ", "").strip()
        if alias not in languages:
            raise InputError(f"Invalid language: `{alias}`")
        if alias not in aliases:
            raise InputError(f"`{alias}` is not an alias")
        print(f"`{alias}` is an alias of `{aliases[alias]}`")
    elif choice.startswith("create alias "):
        choice = choice.replace("create alias ", "").strip()
        split_choice = choice.split()
        if len(split_choice) != 2:
            raise InputError(
                'Error: expected two words after "create alias":\n'
                "the new alias and the language being aliased."
            )
        new_alias, language = split_choice
        if new_alias in aliases:
            if language == aliases[new_alias]:
                raise InputError(f"`{new_alias}` is already an alias of `{language}`.")
            c = input(f"`{new_alias}` is already an alias. Overwrite? (y/n) ")
            if c.lower().split() not in ("y", "yes"):
                raise InputError("Cancelled creating an alias.")
        elif new_alias in languages:
            raise InputError(f"`{new_alias}` is already a language.")
        if language not in languages:
            raise InputError(f"Invalid language: `{language}`")
        if language in aliases:
            language = aliases[language]
        await create_alias(db_file_name, new_alias, language, aliases, languages)
        print(f"Created `{new_alias}` as an alias to `{language}`")
    elif choice.startswith("delete alias "):
        alias = choice.replace("delete alias ", "").strip()
        if alias not in aliases:
            raise InputError(f"`{alias}` is not an alias")
        await delete_alias(alias, aliases, languages, db_file_name)
        print(f"Deleted alias `{alias}`")
    else:
        raise InputError("Invalid input. Enter \x1b[100mhelp\x1b[0m for help.")


async def load_languages(
    loop, session, db_file_name: str, aliases: Dict[str, str]
) -> List[str]:
    """Loads all languages from the database.

    Creates the database with default languages if it doesn't exist. The
    languages include the aliases.
    """
    try:
        with sqlite3.connect(db_file_name) as conn:
            cursor = conn.cursor()
            cursor.execute("SELECT language FROM languages")
            languages: List[Tuple[str]] = cursor.fetchall()
            return [e[0] for e in languages]
    except sqlite3.OperationalError:
        return await create_languages_table(loop, session, db_file_name, aliases)


async def save_languages(db_file_name: str, languages: List[str]) -> None:
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        await _save_languages(cursor, languages)
        conn.commit()


async def _save_languages(cursor, languages: List[str]) -> None:
    languages = [[e] for e in languages]
    cursor.executemany(
        """
        INSERT OR IGNORE INTO languages
        (language)
        VALUES (?);
        """,
        languages,
    )


async def create_languages_table(
    loop, session, db_file_name: str, aliases: Dict[str, str]
) -> List[str]:
    """Creates a database table for and returns all languages.

    Assumes the table does not exist.
    """
    async with await async_tio.Tio(loop=loop, session=session) as tio:
        languages = tio.languages
        languages.extend(aliases.keys())
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            CREATE TABLE languages (
                id INTEGER PRIMARY KEY,
                language TEXT NOT NULL,
                UNIQUE (language)
            );
            """
        )
        await _save_languages(cursor, languages)
        conn.commit()
    return languages


async def print_help() -> None:
    print(
        dedent(
            """\
            help
                Displays this message.
            exit
                Closes this app.
            run \x1b[90;3m(language)\x1b[0m
                Selects a language and then asks you for code to run.
            list
                Shows all supported languages.
            list \x1b[90;3m(search term)\x1b[0m
                Shows all supported languages that contain a chosen search term.
            jargon \x1b[90;3m(language)\x1b[0m
                Shows the code that can wrap around your code in a chosen language.
            create jargon \x1b[90;3m(language)\x1b[0m
                Allows you to set the jargon for a language.
            delete jargon \x1b[90;3m(language)\x1b[0m
                Deletes the jargon for a language.
            alias \x1b[90;3m(alias)\x1b[0m
                Shows the language an alias is an alias of.
            create alias \x1b[90;3m(new alias)\x1b[0m \x1b[90;3m(language)\x1b[0m
                Creates a new alias for a chosen language.
            delete alias \x1b[90;3m(alias)\x1b[0m
                Deletes an alias and any jargon it has.
            For more help, visit https://github.com/wheelercj/tias\
            """
        )
    )


async def list_languages(
    valid_languages: List[str], aliases: Dict[str, str], filter_keyword: str
) -> None:
    """Lists supported languages, optionally filtered with a search term."""
    if filter_keyword:
        valid_languages = list(
            filter(lambda s: filter_keyword in s, valid_languages)
        )
        lang_count = len(valid_languages)
        print(end=f"languages that contain `{filter_keyword}` ({lang_count}): ")
    else:
        print(end=f"languages ({len(valid_languages) - len(aliases)}): ")
    valid_languages = sorted(valid_languages)
    alias_included = False
    for i, language in enumerate(valid_languages):
        if language in aliases:
            alias_included = True
            valid_languages[i] = f"\x1b[36m{language}\x1b[0m"
    valid_languages = ", ".join(valid_languages)
    print(valid_languages)
    if alias_included:
        print("\x1b[90m(Aliases are shown in blue).\x1b[0m")


async def get_code(
    chosen_language: str, aliases: Dict[str, str], db_file_name: str
) -> Tuple[str, str, str]:
    print(end="\x1b[32mcode: \x1b[0m")
    code = get_lines()
    inputs = ""
    if "```" in code:
        code, inputs = await unwrap_code_block(code)
    code = await wrap_jargon(code, chosen_language, db_file_name)
    if chosen_language in aliases:
        chosen_language = aliases[chosen_language]
    return chosen_language, code, inputs


async def run_code(
    loop,
    session,
    db_file_name: str,
    languages: List[str],
    aliases: Dict[str, str],
    chosen_language: str,
    code: str,
    inputs: str,
) -> None:
    async with await async_tio.Tio(loop=loop, session=session) as tio:
        if len(languages) - len(aliases) != len(tio.languages):
            languages = tio.languages
            languages.extend(aliases.keys())
            await save_languages(db_file_name, languages)
        response = await tio.execute(code, language=chosen_language, inputs=inputs)
    print(end=f"\x1b[32m`{chosen_language}` output:\x1b[39m\n{response.stdout}")
    if not response.stdout.endswith("\n"):
        print()
    print(f"\x1b[32mexit status: \x1b[39m{response.exit_status}")


async def unwrap_code_block(statement: str) -> Tuple[str, str]:
    """Removes triple backticks around a code block.

    Returns the input unchanged and an empty string if there are no triple
    backticks. Anything after closing triple backticks is returned as the
    second string. Closing triple backticks are otherwise optional.
    """
    if not statement.startswith("```"):
        return statement, ""
    statement = statement[3:]
    if statement.startswith("\n"):
        statement = statement[1:]
    suffix = ""
    if "```" in statement:
        statement, suffix = statement.split("```", 1)
    if statement.endswith("\n"):
        statement = statement[:-1]
    return statement, suffix


if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print("\nKeyboardInterrupt")



from errors import InputError
from multiline_input import get_lines
from textwrap import dedent
from typing import Dict
from typing import Tuple
import sqlite3


async def init_jargon(db_file_name: str) -> None:
    """Creates a jargon table in the database if it doesn't exist."""
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT name
            FROM sqlite_master
            WHERE type = 'table' AND name = 'jargon';
            """
        )
        if not cursor.fetchall():
            await create_jargon_table(db_file_name)


async def print_jargon(language: str, db_file_name: str) -> None:
    """Shows the jargon for a language if it has jargon."""
    jargon, jargon_key = await load_jargon(language, db_file_name)
    if jargon:
        print(f"\x1b[32mjargon:\x1b[0m\n{jargon}")
        print(f"\x1b[32mjargon key:\x1b[0m {jargon_key}")
    else:
        raise InputError(
            f"No jargon wrapping has been set for the `{language}` language"
        )


async def create_jargon_table(db_file_name: str) -> None:
    """Creates a sqlite table with default jargon.

    Assumes the table does not exist.
    """
    default_jargon: Dict[str, Tuple[str, str]] = await get_default_jargon()
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            CREATE TABLE jargon (
                id INTEGER PRIMARY KEY,
                language TEXT NOT NULL,
                jargon_text TEXT NOT NULL,
                jargon_key TEXT NOT NULL,
                UNIQUE (language)
            );
            """
        )
        for language, (jargon, jargon_key) in default_jargon.items():
            await save_jargon(language, jargon, jargon_key, cursor)
        conn.commit()


async def save_jargon(language: str, jargon: str, jargon_key: str, cursor) -> None:
    """Saves to the database the jargon for a language."""
    cursor.execute(
        """
        INSERT OR IGNORE INTO jargon
        (language, jargon_text, jargon_key)
        VALUES (?, ?, ?);
        """,
        (language, jargon, jargon_key),
    )


async def load_jargon(language: str, db_file_name: str) -> Tuple[str, str]:
    """Gets jargon for a language from the database.

    Returns empty strings if the language has no jargon.

    Returns
    -------
    str
        The language's jargon.
    str
        The language's "jargon key", i.e. something in an expression that, if
        present, means jargon wrapping is not needed.
    """
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT jargon_text, jargon_key
            FROM jargon
            WHERE language = ?;
            """,
            [language],
        )
        records = cursor.fetchall()
        if not records:
            return ("", "")
        jargon, jargon_key = records[0]
        return jargon, jargon_key


async def wrap_jargon(expression: str, language: str, db_file_name: str) -> str:
    """Wraps code around a given expression if the language has needed jargon.

    Returns the expression unchanged if the language has no jargon or if the
    jargon doesn't appear to be needed.
    """
    jargon, jargon_key = await load_jargon(language, db_file_name)
    if not jargon:
        return expression
    if jargon_key not in expression:
        return jargon.replace("INSERT_HERE", expression)
    return expression


async def has_jargon(language: str, db_file_name: str) -> bool:
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            SELECT *
            FROM jargon
            WHERE language = ?;
            """,
            [language],
        )
        records = cursor.fetchall()
        return bool(records)


async def create_jargon(language: str, db_file_name: str) -> None:
    """Gets input for and adds jargon to the database.

    Assumes the language does not already have jargon.
    """
    print("Enter \x1b[100mINSERT_HERE\x1b[0m where you want code to be inserted.")
    print(end="\x1b[32mjargon: \x1b[0m")
    jargon = get_lines()
    if "INSERT_HERE" not in jargon:
        raise InputError("Error: the jargon must contain \x1b[100mINSERT_HERE\x1b[0m")
    print("\x1b[32m---\x1b[0m")
    print("Enter the jargon's key, i.e. a piece of code that, if present")
    print("within code submitted to run, means jargon wrapping is not needed.")
    print(end="\x1b[32mjargon key: \x1b[0m")
    jargon_key = get_lines()
    print("\x1b[32m---\x1b[0m")
    if not jargon_key:
        raise InputError("The jargon key must not be empty")
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            INSERT INTO jargon
            (language, jargon_text, jargon_key)
            VALUES (?, ?, ?);
            """,
            (language, jargon, jargon_key),
        )


async def delete_jargon(language: str, db_file_name: str) -> None:
    with sqlite3.connect(db_file_name) as conn:
        cursor = conn.cursor()
        cursor.execute(
            """
            DELETE FROM jargon
            WHERE language = ?;
            """,
            [language],
        )


async def get_default_jargon() -> Dict[str, Tuple[str, str]]:
    default_jargon: Dict[str, Tuple[str, str]] = {
        # keys: the language
        # values:
        #   * the jargon
        #   * the "jargon key"
        "c": (
            dedent(
                """\
                #include <stdbool.h>
                #include <stdio.h>

                int main(void) {
                    INSERT_HERE
                }\
                """
            ),
            "int main(",
        ),
        "cpp": (
            dedent(
                """\
                #include <iostream>
                #include <stdio.h>
                using namespace std;

                int main() {
                    INSERT_HERE
                }\
                """,
            ),
            "int main(",
        ),
        "cs": (
            dedent(
                """\
                namespace MyNamespace {
                    class MyClass {         
                        static void Main(string[] args) {
                            INSERT_HERE
                        }
                    }
                }\
                """
            ),
            "static void Main(",
        ),
        "dart": (
            dedent(
                """\
                void main() {
                    INSERT_HERE
                }\
                """
            ),
            "void main(",
        ),
        "go": (
            dedent(
                """\
                package main
                import "fmt"

                func main() {
                    INSERT_HERE
                }\
                """
            ),
            "func main(",
        ),
        "java": (
            dedent(
                """\
                import java.util.*;

                class MyClass {
                    public static void main(String[] args) {
                        Scanner scanner = new Scanner(System.in);
                        INSERT_HERE
                    }
                }\
                """
            ),
            "public static void main(",
        ),
        "kotlin": (
            dedent(
                """\
                fun main(args : Array<String>) {
                    INSERT_HERE
                }\
                """
            ),
            "fun main(",
        ),
        "objective-c": (
            dedent(
                """\
                #include <stdio.h>
                // Print with the `puts` function, not `NSLog`.
                
                int main() {
                    INSERT_HERE
                }\
                """
            ),
            "int main(",
        ),
        "rust": (
            dedent(
                """\
                fn main() {
                    INSERT_HERE
                }\
                """
            ),
            "fn main(",
        ),
        "scala": (
            dedent(
                """\
                object Main extends App {
                    INSERT_HERE
                }\
                """
            ),
            "object Main",
        ),
    }
    default_jargon["c-clang"] = default_jargon["c"]
    default_jargon["c-gcc"] = default_jargon["c"]
    default_jargon["c-tcc"] = default_jargon["c"]
    default_jargon["c#"] = default_jargon["cs"]
    default_jargon["c++"] = default_jargon["cpp"]
    default_jargon["cpp-clang"] = default_jargon["cpp"]
    default_jargon["cpp-gcc"] = default_jargon["cpp"]
    default_jargon["cs-core"] = default_jargon["cs"]
    default_jargon["cs-csc"] = default_jargon["cs"]
    default_jargon["cs-csi"] = default_jargon["cs"]
    default_jargon["cs-mono-shell"] = default_jargon["cs"]
    default_jargon["cs-mono"] = default_jargon["cs"]
    default_jargon["java-jdk"] = default_jargon["java"]
    default_jargon["java-openjdk"] = default_jargon["java"]
    default_jargon["objective-c-clang"] = default_jargon["objective-c"]
    default_jargon["objective-c-gcc"] = default_jargon["objective-c"]
    return default_jargon
